#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <Hash.h>
#include <ESPAsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ESP8266WiFi.h>
#include <Wire.h>
#include <PubSubClient.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>
#include "LiquidCrystal.h"

const int RS = 16, EN = 0, d4 = 12 , d5 = 13, d6 = 15, d7 = 3;   // display 1602 parameters (define as per your display) see end of the file for my detailed settings

LiquidCrystal lcd(RS, EN, d4, d5, d6, d7);   //  display 1602 parameters (define as per your display) see end of the file for my detailed settings

#define SEALEVELPRESSURE_HPA (1013.25)

#define SETPOINT 42   // desired humidity 
#define DEADBAND 2    // parameters to maintain humidity

int DEHYDRATOR = D5; // pin of arduino/ESP to which relay SIG terminal is connected
#define ON true      // Energise the coil and close the circuit for NO relay
#define OFF false    // 


#define ssid "wifissid "   // change to your Router ssid
#define password "wifipassword"  // your ssid password

#define mqtt_server "192.168.1.236"   //define your mqtt server address
#define mqtt_user "mqtt"
#define mqtt_password "****"   // define password for mqtt server 

WiFiClient espClient;
PubSubClient client(espClient);
long lastMsg = 0;
char msg[50];

void setup_wifi() {
  delay(100);
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);

  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) 
  {
    delay(500);
    Serial.print(".");
  }
  randomSeed(micros());
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
}

void callback(char* topic, byte* payload, unsigned int length) 
{
  
} //end callback

void reconnect() {
  // Loop until we're reconnected
  while (!client.connected()) 
  {
    Serial.print("Attempting MQTT connection...");
    String clientId = "ESP8266Client-";
    clientId += String(random(0xffff), HEX);
    // Attempt to connect
    // If you do not want to use a username and password, change next line to
    // if (client.connect("ESP8266Client")) {
    if (client.connect("ESP8266Client", mqtt_user, mqtt_password)) {
      Serial.println("connected");
      client.subscribe("dehydrator");
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
}

float t = 0.0;
float h = 0.0;
float p = 0.0;

AsyncWebServer server(80);

const long interval = 1000; 

unsigned long previousMillis = 0;

// In case you have more than one sensor, make each one a different number here  
#define sensor_number "1"
#define humidity_topic "sensor/humidity"                   // mqtt topics
#define temperature_c_topic "sensor/temperaturec"          // mqtt topics
#define temperature_f_topic "sensor/temperaturef"          // mqtt topics
#define barometer_hpa_topic "sensor/pressurehpa"           // mqtt topics
#define barometer_inhg_topic "sensor/pressureinhg"         // mqtt topics
// Lookup for your altitude and fill in here, units hPa
// Positive for altitude above sea level
#define baro_corr_hpa 34.5879 // = 289m above sea level

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html>
<head>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="https://use.fontawesome.com/releases/v5.7.2/css/all.css" integrity="sha384-fnmOCqbTlWIlj8LyTjo7mOUStjsKC4pOpQbqyi7RrhN7udi9RwhKkMHpvLbHG9Sr" crossorigin="anonymous">
  <style>
    html {
     font-family: Arial;
     display: inline-block;
     margin: 0px auto;
     text-align: center;
    }
    h2 { font-size: 3.0rem; }
    p { font-size: 3.0rem; }
    .units { font-size: 1.2rem; }
    .bme-labels{
      font-size: 1.5rem;
      vertical-align:middle;
      padding-bottom: 15px;
    }
  </style>
</head>
<body>
  <h2>ESP8266 BME Server</h2>
  <p>
    <i class="fas fa-thermometer-half" style="color:#ffa500;"></i> 
    <span class="bme-labels">Temperature</span> 
    <span id="temperature">%TEMPERATURE%</span>
    <sup class="units">&deg;C</sup>
  </p>
  <p>
    <i class="fas fa-tint" style="color:#00add6;"></i> 
    <span class="bme-labels">Humidity</span>
    <span id="humidity">%HUMIDITY%</span>
    <sup class="units">&#37;</sup>
  </p>
  <p>
    <i class="fas fa-tachometer-alt" style="color:#059e8a;"></i> 
    <span class="bme-labels">Pressure</span>
    <span id="pressure">%PRESSURE%</span>
    <sup class="units">hpa</sup>
  </p>
</body>
<script>
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("temperature").innerHTML = this.responseText;
    }
  };
  xhttp.open("GET", "/temperature", true);
  xhttp.send();
}, 10000 ) ;

setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("humidity").innerHTML = this.responseText;
    }
  };
  xhttp.open("GET", "/humidity", true);
  xhttp.send();
}, 10000 ) ;

setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("pressure").innerHTML = this.responseText;
    }
  };
  xhttp.open("GET", "/pressure", true);
  xhttp.send();
}, 10000 ) ;

</script>
</html>)rawliteral";

// Replaces placeholder with BME values
String processor(const String& var){
  //Serial.println(var);
  if(var == "TEMPERATURE"){
    return String(t);
  }
  else if(var == "HUMIDITY"){
    return String(h);
  }
  return String();
}


Adafruit_BME280 bme; // I2C
Adafruit_Sensor *bme_temp = bme.getTemperatureSensor();
Adafruit_Sensor *bme_pressure = bme.getPressureSensor();
Adafruit_Sensor *bme_humidity = bme.getHumiditySensor();

void setup() {
   Serial.begin(115200);
   setup_wifi();
   client.setServer(mqtt_server, 1883);
   client.setCallback(callback);
  // Setup the two LED ports to use for signaling status
  pinMode(DEHYDRATOR, OUTPUT);   // DEHYDRATOR - see DEHYDRATOR = D5; above
  digitalWrite(DEHYDRATOR, OFF);
   
 
  WiFi.begin(ssid, password);
  

  // Set SDA and SDL ports
  //Wire.begin(2, 14);
  // Using I2C on the HUZZAH board SCK=#5, SDI=#4 by default

  // Start sensor
  if (!bme.begin(0x76)) {        // Some BME sensors have 0x77 as default port eg Waveshare - change accordingly
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    while (1);
  }
  lcd.begin(16,2);         // 1602 display
  lcd.clear();

  lcd.setCursor(0,0);
   lcd.print("Temp:");

   lcd.setCursor(0,1);
    lcd.print("Humidity:");
   // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html, processor);
  });
  server.on("/temperature", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", String(t).c_str());
  });
  server.on("/humidity", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", String(h).c_str());
  });
  server.on("/pressure", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/plain", String(p).c_str());
  });

  // Start server
  server.begin(); 
  
}





bool checkBound(float newValue, float prevValue, float maxDiff) {
  return newValue < prevValue - maxDiff || newValue > prevValue + maxDiff;
}


long lastForceMsg = 0;
bool forceMsg = false;
float temp = 0.0;
float hum = 0.0;
float baro = 0.0;
float diff = 1.0;


void loop() {
  delay(5000);
  float newH = bme.readHumidity();
  float newT = bme.readTemperature();

  lcd.setCursor(6, 0);
  lcd.print(newT,2); 

  lcd.setCursor(10, 1);
   lcd.print(newH,2);

  if(digitalRead(DEHYDRATOR) == OFF)
   {
    //If dehydrator is OFF check to see if humidity is above setpoint
    if( h > SETPOINT + DEADBAND)
    {
      //We have reached the higher limit, so start the dehydrator
      digitalWrite(DEHYDRATOR, ON);  
    }
   }
   else 
   {
    //if dehydrator is ON check to see if humidity is below setpoint
    if( h < SETPOINT - DEADBAND)
    { //we have reached the lower limit, so kick off the dehydrator
     digitalWrite(DEHYDRATOR, OFF); 
    }
   } 

  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis >= interval) {
    // save the last time you updated the BME values
    previousMillis = currentMillis;
    // Read temperature as Celsius (the default)
    float newT = bme.readTemperature();
    // Read temperature as Fahrenheit (isFahrenheit = true)
    //float t = dht.readTemperature(true);
    // if temperature read failed, don't change t value
    if (isnan(newT)) {
      Serial.println("Failed to read from BME sensor!");
    }
    else {
      t = newT;
      Serial.println(t);
    }
    // Read Humidity
    float newH = bme.readHumidity();
    // if humidity read failed, don't change h value 
    if (isnan(newH)) {
      Serial.println("Failed to read from BME sensor!");
    }
    else {
      h=newH;
      Serial.println(h);
    }
     float newP = bme.readPressure() / 100.0F;
    // if humidity read failed, don't change h value 
    if (isnan(newP)) {
      Serial.println("Failed to read from BME sensor!");
    }
    else {
      p=newP;
      Serial.println(p);
    }
  }   
  
  if (!client.connected()) {
    reconnect();
  }
  client.loop();

  long now = millis();
  int status;
  if (now - lastMsg > 2000) {
    lastMsg = now;
    status=digitalRead(DEHYDRATOR);
     String msg="Button status: ";
     if(status==HIGH )
    {
        msg= msg+ "Pressed";     // message mqtt broker about dehydrator status 
       char message[58];
       msg.toCharArray(message,58);
       Serial.println(message);
       //publish sensor data to MQTT broker
      client.publish("dehydrator", message);
       }
     else
     {
      msg= msg+ " Not Press";
       char message[58];
       msg.toCharArray(message,58);
       Serial.println(message);
       //publish sensor data to MQTT broker
      client.publish("dehydrator", message);
     }
    // MQTT broker could go away and come back at any time
    // so doing a forced publish to make sure something shows up
    // within the first 5 minutes after a reset
    if (now - lastForceMsg > 300000) {
      lastForceMsg = now;
      forceMsg = true;
      Serial.println("Forcing publish every 5 minutes...");
    }

    float newTemp = bme.readTemperature();
    float newHum = bme.readHumidity();
    float newBaro = bme.readPressure() / 100.0F;

    if (checkBound(newTemp, temp, diff) || forceMsg) {
      temp = newTemp;
      float temp_c=temp; // Celsius
      float temp_f=temp*1.8F+32.0F; // Fahrenheit
      Serial.print("New temperature:");
      Serial.print(String(temp_c) + " degC   ");
      Serial.println(String(temp_f) + " degF");
      client.publish(temperature_c_topic, String(temp_c).c_str(), true);
      client.publish(temperature_f_topic, String(temp_f).c_str(), true);
    }

    if (checkBound(newHum, hum, diff) || forceMsg) {
      hum = newHum;
      Serial.print("New humidity:");
      Serial.println(String(hum) + " %");
      client.publish(humidity_topic, String(hum).c_str(), true);
    }

    if (checkBound(newBaro, baro, diff) || forceMsg) {
      baro = newBaro;
      float baro_hpa=baro; // hPa corrected to sea level
      float baro_inhg=(baro)/33.8639F; // inHg corrected to sea level
      Serial.print("New barometer:");
      Serial.print(String(baro_hpa) + " hPa   ");
      Serial.println(String(baro_inhg) + " inHg");
      client.publish(barometer_hpa_topic, String(baro_hpa).c_str(), true);
      client.publish(barometer_inhg_topic, String(baro_inhg).c_str(), true);
    }

    forceMsg = false;
  }
}


/*******   
 display connections  left side display pins >> Right side ESP pins 
 VSS >> GND
 VDD >> 3V3
 V0 >> GND
 RS>> D0
 RW >> GND
 E >> D3
 D4 (Display) >>D6 (ESP)
 D5 (display) >> D7 (ESP)
 D6 (display) >> D8 (ESP)
 D7 (display) >> RX (ESP)
 A (display) >> 3V3
 K >> GND 

 BME connections   Left side BME >> Right side ESP

 VCC >> 3V3
 GND
 SDA >> D2
 SCL >> D1

 Relay             Left side Relay >> Right side ESP

 VCC>> 3V3
 GND
 SIG >> D5
     
ESP pinout display 
static const uint8_t D0   = 16;
static const uint8_t D1   = 5;
static const uint8_t D2   = 4;
static const uint8_t D3   = 0;
static const uint8_t D4   = 2;
static const uint8_t D5   = 14;
static const uint8_t D6   = 12;
static const uint8_t D7   = 13;
static const uint8_t D8   = 15;
static const uint8_t D9   = 3;
static const uint8_t D10  = 1;
   
******/
